#ifndef STACK_THEME_H
#define STACK_THEME_H

#include "theme.h"

namespace OCC {

class StackTheme : public Theme
{

public:
    StackTheme() {};

    bool linkSharing() const override {
        return true;
    }

    bool userGroupSharing() const override {
        return false;
    }

    bool showVirtualFilesOption() const override {
        return true;
    }

    QColor wizardHeaderBackgroundColor() const override {
        // STACK color #5252c5
        return QColor(82, 82, 197);
    }

    QColor wizardHeaderTitleColor() const override {
        return QColor(255, 255, 255);
    }

    // Note: Disabling custom button colours until we find some that actually look better than stock

    // QmlButtonColor primaryButtonColor() const override {
    //     const QColor button("#5252c5");
    //     return {button, Qt::white, button.darker()};
    // }

    // QmlButtonColor secondaryButtonColor() const override {
    //     return {"#cae4ff", Qt::black, QColor(Qt::black).lighter()};
    // }

    bool systrayUseMonoIcons() const {
        return true;
    }

    bool aboutShowCopyright() const override {
        return false;
    }

    QIcon wizardHeaderLogo() const override {
        return Resources::themeUniversalIcon(QStringLiteral("wizard_logo"));
    }

    QIcon aboutIcon() const override {
        return Resources::themeUniversalIcon(QStringLiteral("stack-icon"));
    }

    QString helpUrl() const override {
        return QStringLiteral("https://stackstorage.app");
    }

    QString defaultClientFolder() const {
        return QStringLiteral("STACK");
    }

    QString about() const override {
        return tr(
            "<h2>STACK client</h2>"
            "<p>Versie %1. Zie <a href='%2'>%3</a> voor meer informatie.</p>"
            "<p>Deze release wordt je aangeboden door team.blue.</p>"
            "<hr>"
            "<p><small>By Klaas Freitag, Daniel Molkentin, Olivier Goffart, Markus Götz,<br/>"
            "&nbsp;&nbsp;&nbsp; Jan-Christoph Borchardt, Thomas Müller, Dominik Schmidt, Michael Stingl, Hannah von Reth, and others.</small></p>"
            "<p>Copyright ownCloud GmbH</p>"
            "<p>Licensed under the GNU General Public License (GPL) Version 2.0.<br/>"
            "%5 and the %5 logo are registered trademarks of %5 in the <br/>"
            "United States, other countries, or both.</p>"
        )
        .arg(
            Utility::escape(OCC::Version::displayString()),
            Utility::escape(QStringLiteral("https://" APPLICATION_DOMAIN)),
            Utility::escape(QStringLiteral(APPLICATION_DOMAIN)),
            Utility::escape(QStringLiteral("ownCloud GmbH"))
        );
    }
    
    bool moveToTrashDefaultValue() const override {
        // The move-to-trash option is enabled by default
        return true;
    }

    bool spacesAreCalledFolders() const override {
        return true;
    }

    ;

private:
};
}
#endif
