/*
 * Copyright (C) by Hannah von Reth <hannah.vonreth@owncloud.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#pragma once

#include "libsync/networkjobs/jsonjob.h"

#include <QObject>

namespace OCC {
class Folder;
/*
 * Migration for cern eos
 * This class enabled migration from legacy dav urls to native eos paths.
 *  host/remote.php/dav/files/Einstein/eos/e -> host/.../eos/e
 */
class SpaceMigration : public QObject
{
    Q_OBJECT
public:
    SpaceMigration(const AccountStatePtr &accountState, const QString &path, QObject *parent);
    void start();

Q_SIGNALS:
    void finished();

private:
    void migrate(const QJsonObject &folders);

    AccountStatePtr _accountState;
    const QString _path;

    QVector<QPointer<Folder>> _migrationFolders;
};

}
